#!/usr/bin/perl
#
#	Package containing the functional forms of the power-law computation.
#
#	Revision:	$Revision: 1.6 $
#	Date:	$Date: 2016/08/30 10:40:34 $
#--------------------------------------------------------------
#
package	GEN::Power;

use	strict;
use	warnings;

use	Exporter	qw(import);

use	constant	STERLING1 => 1;
use	constant	STERLING2 => 2;

our	@EXPORT_OK	= qw(sterling power gammaln eulersum modinfo STERLING1 STERLING2);

#--------------------------------------------------------
sub	gammaln
{
#	Returns the log of the gamma function for its single argument > 0.
#	Taken from Press et. al., "Numerical Recipes in C" p. 214, 2nd edition 
#	and transcribed to perl.
#
#	Valid for reals.  Working with logs controls overflow.  This is very useful
#	and is closely related to Stirling's formula for integers.
#
#	Example: my $val = exp(gammaln(5.0));
#	returns the value 4.3.2.1 = 24 (since gamma(n+1) = n!).
#
	my	($xx)	= @_;

	my	@cof		= (
					76.18009172947146,
					-86.50532032941677,
					24.01409824083091,
					-1.231739572450155,
					0.1208650973866179e-2,
					-0.5395239384953e-5
				);

	my	$y		= $xx;	
	my	$x		= $xx;	

	my	$tmp		= $x + 5.5;
	$tmp		-= ($x + 0.5) * log($tmp);

	my	$ser		= 1.000000000190015;
	for( my $j = 0; $j <= 5; ++$j )		{ $ser	+= $cof[$j]/(++$y); }

	my	$result	= -$tmp + log((2.5066282746310005 * $ser)/$x);

#	print "gammaln: xx = $xx, result = $result\n";

	return	-$tmp + log((2.5066282746310005 * $ser)/$x);
}

#--------------------------------------------------------
sub	eulersum
{
#
#	Computes the Euler sum component of the information modulating function modfunc()
#	in the paper.
#
	my	($ti,$ai)	= @_;
	my	$n		= 30;
	my	$sum		= 0.0;

	for( my $i = 1; $i <= $n; ++$i )
	{
		$sum	+= 1.0 / (($ti-$ai+1+$n)*($ti+1+$n));
	}

	return	- $ai * $sum;
}

#--------------------------------------------------------
sub	sterling
{
#
#	Computes Sterling's first and second order adjustments to his
#	factorial approximation.
#
	my	($ti,$approx)	= @_;
#
#	Next order is:-
#	exp((-0.5/$ti) + (1.0/(12.0*$ti*$ti)));
#
	if ( $approx == STERLING1 )		{ return exp(-0.5/$ti); }
	elsif ( $approx == STERLING2 )	{ return exp((-0.5/$ti) + (1.0/(12.0*$ti*$ti))); }
	else							{ die "Illegal value of approx in sterling().\n"; }
}

#--------------------------------------------------------
sub	modinfo
{
	my	($ti,$ai,$beta)	= @_;
#
#	The Stirling approximation is only true for n > 0 since log n
#	occurs and our expansion includes log(ti-ai).
#
	if ( $ti < $ai )		{ return 1.0; }

	my	$fact			= ($ai/(($ti-$ai+1)*($ti+1))) - eulersum($ti,$ai);
	my	$mfunc			= exp(-($beta*$fact));

#	print "($ti,$ai,$beta): mfunc = $mfunc, first = $first, numerator = $numerator, denominator = $denominator\n";

	return	$mfunc;
}

#--------------------------------------------------------
sub	invpower
{
#
#	Returns the inverse power law value.
#
	my	($ti,$beta,$scalar)	= @_;
	return	($scalar/$ti)**((1.0/$beta));
}
#--------------------------------------------------------
sub	power
{
#
#	Returns the power law value.
#
	my	($ai,$beta,$scalar)	= @_;
	return	($scalar*(($ai)**(-$beta)));
}
#--------------------------------------------------------
#
#	Standard exit to return 1 for eval when called.
#
1;
