#!/usr/bin/perl
#
#	Calculate the N(ti,ai;ai) function by recursion.
#
#	Revision:	$Revision: 1.4 $
#	Date:	$Date: 2017/11/01 15:11:43 $
#--------------------------------------------------------------
#
use 	strict;
use 	warnings;
use	Getopt::Long;

my	$BETA	= 3.0;
my	$ALPHA	= 6.0;
my	$TIMIN	= 1;
my	$TIMAX	= 300;

my	%opt		= ();

my	@pdfaipwr	= ();
my	@pdfai	= ();
my	@pdfaid	= ();
my	@pdfti	= ();

GetOptions(
	'alpha=f'		=> \$ALPHA,
	'beta=f'		=> \$BETA,
	'd'    		=> \$opt{d},			# Generate defect model
	'f'    		=> \$opt{f},			# Generate normal full distribution
	'p'    		=> \$opt{p},
	's'    		=> \$opt{s},
	'timin=i'     	=> \$TIMIN,
	'timax=i'     	=> \$TIMAX,
	'v'     		=> \$opt{v},
	'verbose'     	=> \$opt{v},
);
#
#	General purpose integer factorial function.
#
sub	factorial()
{
	my	($n)		= @_;

	my	$result	= 1;
	for( my $i = 2; $i <= $n; ++$i )
	{
		$result	*= $i;
	}

	return	$result;
}
#
#	The nCr operator.
#
sub	nCr
{
	my	($n,$r)	= @_;

	my	$result	= &factorial($n) / (&factorial($r) * &factorial($n-$r));

	return	$result;	
}
#
#	The lower bound for N(ti,ai;ai), (ti!)/((ti/ai)!)^(ai)
#	using integer decomposition.
#
sub	calcLowerN
{
	my	($ti,$ai)	= @_;

	my	$remainder	= $ti % $ai;
	my	$m			= ($ti - $remainder) / $ai;
	my	$factm		= &factorial($m);
	my	$result		= &factorial($ti) / (($factm**($ai-1)) * &factorial($m+$remainder));

	return	$result;
}
#
#	Calculates the full N(ti,ai;ai) using recursion as described in the arXiv paper.
#
sub	calcFullN
{
	my	($ti,$ai)	= @_;

	my	%elements	= ();

	$elements{1 . "-" . 1}	= 1;

	for( my $i = 2; $i <= $ai; ++$i )
	{
		for( my $j = 1; $j < $i; ++$j )
		{
			$elements{$i . "-" . $j}	= &nCr($i,$j) * $elements{$j . "-" . $j};
			my	$value		= $elements{$i . "-" . $j};
			print "---> N($ti,$i;$j) = nCr($ai,$j) * N($ti,$j;$j) = $value\n" if $opt{v};
		}

		$elements{$i . "-" . $i}	= ($i**$ti);
#		print "N($ti,$i;$i) = ($i**$ti)\n";
		for( my $j = 1; $j < $i; ++$j )
		{
			$elements{$i . "-" . $i}	-= $elements{$i . "-" . $j};
			my	$value		= $elements{$i . "-" . $j};
			print " - N($ti,$i;$j) ($value)\n" if $opt{v};
		}
	}

	print "###> N($ti,$ai;$ai) = $elements{$ai . '-' . $ai};\n" if $opt{v};

	return	$elements{$ai . "-" . $ai};
}
#
#	It overflows after 300.
#
my	%N			= ();
my	%dlogNdti		= ();
#
#	Populate the N() data structure.
#
my	$TIMAXFULL	= 300;	# Overflow beyond this.

for( my $ti = $TIMIN; $ti <= $TIMAXFULL; ++$ti )
{
	my	$maxai	= $ti;
	if ( $ti > 30 )		{ $maxai	= 30; }

	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	$value		= &calcFullN($ti, $ai);

		$N{$ai . "-" . $ti}	= $value;
	}
}
#
#	Calculate the derivatives.
#

my	$aisum	= 0.0;
my	$firstti	= 0;
my	$firstaii	= 0;
my	$lastti	= 0;

for( my $ti = $TIMIN; $ti <= $TIMAXFULL; ++$ti )
{
#
#	Incorporates Ramanujan's approximation for log n!.
#	https://en.wikipedia.org/wiki/Stirling%27s_approximation
#	Setting it equal to 0 gives the derivative of the simplest Stirling approximation d/dn (n*ln(n) - n) = log n.
#	Ramanujan's approximation is:-
#
#		log n! \approx n log n - n + 1/6 log(n + 4n^2 + 8n^3) + 1/2log(\pi)
#
#	Taking the derivative of this gives the srinivasa term below.
#
#	my	$srinivasa	= 0.0;
	my	$srinivasa	= (1.0 + 8*$ti + 24*$ti*$ti) / (6.0*($ti + 4*$ti*$ti + 8*$ti*$ti*$ti));

	my	$lhs			= -(log($ti) + $srinivasa - $ALPHA)/$BETA;
	my	$tip1		= $ti + 1;
	my	$tim1		= $ti - 1;

	if ( defined $N{$ti . "-" . $ti} && $N{$ti . "-" . $ti} != 0 && $ti < 20 )
	{
#
#		We calculate the "exact" value of lhs using d/dti log ti! just for comparison.
#
		my	$exact	= -(log(&factorial($tip1)) - $ALPHA*$tip1)/$BETA + (log(&factorial($ti)) - $ALPHA*$ti)/$BETA;
		my	$logN	= log $N{$ti . "-" . $ti};
		print ":::> ti = $ti, N{$ti . '-' . $ti} = $N{$ti . '-' . $ti}, lhs = $lhs, exact = $exact, log N = $logN\n" if $opt{v};
	}

	my	$maxai	= $ti;
	if ( $ti > 30 )		{ $maxai	= 30; }

	my	$apihigh	= 1;
	my	$apilow	= $maxai;
	my	$straddle	= 0;

	my	%exprhs		= ();
	my	%explhs		= ();

	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	$aip1	= $ai + 1;

		if ( defined $N{$ai . "-" . $tim1} && defined $N{$ai . "-" . $tip1} )
		{
#
#			Centred difference.
#
			$dlogNdti{$ai . "-" . $ti}	= 0.5 * (log($N{$ai . "-" . $tip1}) - log($N{$ai . "-" . $tim1}));
		}
		else
		{
			$dlogNdti{$ai . "-" . $ti}    = 0.0;
		}

		my	$dlogaitidti	= log($ai**$tip1) - log($ai**$ti);

		my	$rhspower		= $dlogaitidti;
		my	$rhsfull		= $dlogNdti{$ai . "-" . $ti};
		my	$aipower		= exp($lhs);
		my	$exprhspower	= exp($rhspower);
		my	$exprhsfull	= exp($rhsfull);

		$exprhs{$ai}		= $exprhsfull;
		$explhs{$ai}		= $aipower;

		if ( $exprhsfull < $aipower )			{ $apilow	= $exprhsfull; }
		if ( $aipower < $exprhsfull )			{ $apilow	= $exprhsfull; }
#
#		Finally, calculate the lower bound.
#
		my	$lowerN		= &calcLowerN($ti,$ai);
		my	$fullN		= $N{$ai . "-" . $ti};
		my	$powerN		= $ai**$ti;

		my	$loglowerN	= log($lowerN);
		my	$logfullN		= log($fullN);
		my	$logpowerN	= log($powerN);
#
#		Various diagnostics.
#
		print "===> ti = $ti, ai = $ai, lhs = $lhs, rhs = $rhsfull, exprhsfull = $exprhsfull\n" if $opt{v};
#		print "     loglowerN = $loglowerN, logfullN = $logfullN, logpowerN = $logpowerN\n" if $opt{v};
	}
#
#	Now interpolate.
#
#	aipower is the value of ai which is the solution of the pure power-law equation.
#
#	To find the equivalent solution for the full equation, we must find which value of ai
#	forces N(ti,ai;ai) = aipower.
#
	for( my $ai = 1; $ai <= $maxai; ++$ai )
	{
		my	$aip1	= $ai + 1;

		if ( defined $exprhs{$ai} && defined $exprhs{$aip1} )
		{
			if ( ($exprhs{$ai} < $explhs{$ai}) && ($exprhs{$aip1} >= $explhs{$ai}) )
			{
#
#				Straddle.
#
				my	$ai_interp	= $ai + ($explhs{$ai} - $exprhs{$ai})/($exprhs{$aip1} - $exprhs{$ai});
			$dlogNdti{$ai . "-" . $ti}	= 0.5 * (log($N{$ai . "-" . $tip1}) - log($N{$ai . "-" . $tim1}));
				my	$dlogN_interp	= $dlogNdti{$ai . "-" . $ti} + ($dlogNdti{$aip1 . "-" . $ti} - $dlogNdti{$ai . "-" . $ti})*($ai_interp - $ai)/($aip1 - $ai);
				my	$tiaidiff		= $ti - $ai_interp;

				print "***> ti = $ti, ai = $ai, aipower = $explhs{$ai}, exprhs{$ai} = $exprhs{$ai}, exprhs{$aip1} = $exprhs{$aip1}, ai_interp = $ai_interp, tiaidiff = $tiaidiff\n" if $opt{v};
#				print "$ti $explhs{$ai} $ai_interp\n";

				if ( $firstti == 0 )
				{
#
#					Set the first ti for which we have a solution.
#					We currently do not use this.  It could be used to interpolate the low
#					end but this is artificial.
#
					$firstti	= $ti;
					$firstaii	= $ai_interp;
				}

				$lastti	= $ti;

				push @pdfaipwr, $explhs{$ai};
				push @pdfai, $ai_interp;
				push @pdfaid, $dlogN_interp;
				push @pdfti, $ti;

				$aisum	+= $ai_interp;
			}
			else
			{
				print "+++> ti = $ti, ai = $ai, aipower = $explhs{$ai}, exprhs{$ai} = $exprhs{$ai}, exprhs{$aip1} = $exprhs{$aip1}\n" if $opt{v};
			}
		}
	}
}
#
#	Fill out the rest with the power-law as it is by now indistinguishable.
#
for( my $ti = $lastti+1; $ti <= $TIMAX; ++$ti )
{
	my	$lhs			= -(log($ti) - $ALPHA)/$BETA;
	my	$aipower		= exp($lhs);

#	print "$ti $aipower $aipower\n";

	push @pdfaipwr, $aipower;
	push @pdfai, $aipower;
	push @pdfaid, ($ti * log($aipower));
	push @pdfti, $ti;

	$aisum	+= $aipower;
}

while(1)
{
	my	$ti		= shift(@pdfti);
	my	$ai		= shift(@pdfai);
	my	$di		= $ti * shift(@pdfaid);
	my	$aipwr	= shift(@pdfaipwr);
	my	$aibc	= $ti;

	if ( ! defined $ti )		{ last; }

	if ( $opt{p} )
	{
		$ai		/= $aisum;
		$aipwr	/= $aisum;
		$aibc	/= $aisum;
	}

	print "$ti $aipwr $ai $aibc\n" if $opt{f};
	print "$ti $di\n" if $opt{d};
}

if ( $opt{s} )
{
	print "\n\n";

	for ( my $ti = 0; $ti < 100; ++$ti )
	{
		my	$aibc	= $ti / $aisum;
		print "$ti $aibc\n";
	}
}

exit(0);
