#!/usr/bin/perl
#
#	Solve the implicit relation for (ti,ai) pairs.
#	Notice that we do this whether the relation is implicit or
#	explicit.
#
#	Revision:	$Revision: 1.1 $
#	Date:	$Date: 2017/01/21 16:19:11 $
#--------------------------------------------------------------
#
use	strict;
use	warnings;
use	Getopt::Long;
#
#	Tell perl how to find our home-brewed package.  This is called Power.pm and
#	is located under ./GEN, (we are in GENOMICS here).
#
use	File::Basename qw(dirname);
use	Cwd	qw(abs_path);
use	lib	dirname(dirname abs_path $0) . '/GENOMICS';
use	GEN::Power qw(power sterling modinfo STERLING1 STERLING2);

my	%opt	= ();
#
#	f 	- Use full algorithm; otherwise asymptotic algorithm.
#	a	- find pdf of ai
#	t	- find pdf of ti
#
my	$scalar		= 100000;
my	@ai_out		= ();
my	@ti_out		= ();
my	$ai			= 0.0;
my	$beta		= 1.85;
my	$fixa		= 0.0;
my	$timin		= 5;
my	$timax		= 2000;

my	$aimin		= 0.15;		# Was 5.  This gets convergence for large ti.
my	$aimax		= 1000.0;		# Was 100, but this is crucial for convergence around
							# the peak.  The ti >= ai requirement guarantees relevant solutions.
							# You have to let the bisection algorithm wander outside the
							# realm of feasible solution to find a feasible solution.


GetOptions(
	'a'        	=> \$opt{a},
	'beta=f'		=> \$beta,
	'conv'		=> \$opt{c},
	'f'        	=> \$opt{f},
	'fixa=f'		=> \$fixa,
	'full'        	=> \$opt{f},
	'n'        	=> \$opt{n},
	'nobc'		=> \$opt{b},
	'scalar=i'	=> \$scalar,
	't'        	=> \$opt{t},
	'timax=i'     	=> \$timax,
	'timin=i'     	=> \$timin,
	'v'     		=> \$opt{v},
	'vv'     		=> \$opt{vv},
	'vvv'    		=> \$opt{vvv},
	'verbose'     	=> \$opt{vv},
);

if ( $opt{a} && $opt{t} )					{ die "Can't have both -a and -t\n"; }
if ( $fixa > 0.0 && ($opt{a} || $opt{t}) )		{ die "-fixa and (-a or -t) interfere\n"; }
#--------------------------------------------------
#	Use the forward model to generate a value to put in the full
#	equation.  When result == 0, we have a match.
#
#	The forward model is ti = $sfunc * $mfunc * $ai^(-$beta)
#
sub	generateTIResidual()
{
	my	($ti,$ai,$scalar,$beta, $sfact)	= @_;
	my	$ti_estimate_full				= 0.0;
	my	$ti_estimate_asymp				= 0.0;
	my	$pfunc						= power($ai,$beta,$scalar);
	my	$sfunc						= sterling($ti,STERLING1);
	my	$mfunc						= modinfo($ti,$ai,$beta);
#
#	Use full implicit equation and asymptotic version.
#
#	Holding ai constant, we want to find
#		delta-ti (log ti! - log (ti-ai)! + (ti-ai) log ai) to minimise
#	the log Omega expression.  Using the gamma function, we are minimising
#		delta-ti ( log gamma(ti+1) - log gamma(ti-ai+1) + (ti-ai) log ai)
#	However, this is given by
#		delta-ti d/dti ( log gamma(ti+1) - log gamma(ti-ai+1) + (ti-ai) log ai)
#	Now the derivative of the log of the gamma function has a very pretty form.
#		d/dti log gamma(ti) = -C - 1/ti + sum_n=1^inf (ti/n(ti+n)
#	Substitute in the above and the following simplification emerges
#		delta-ti (ai/((ti-ai+1)(ti+1)) + log ai + sum ...)
#
#	Note that this also has the right asymptotic form with the first term
#	tending to 0 as ai/ti -> 0
#
	$ti_estimate_full	= $sfunc * $mfunc * $pfunc;
	$ti_estimate_asymp	= $sfunc * $pfunc;

	my	$result				= 0.0;

	if ( $ai > 0.1 )
	{
		if ( $opt{f} )			{ $result   = $ti - $ti_estimate_full; }
		else					{ $result   = $ti - $ti_estimate_asymp; }
	}

	if ( $opt{v} )
	{
		printf "(ti,ai) = (%5.2f,%5.2f) ti_estimate_full = %15.7f, ti_estimate_asymp = %15.7f\n", 
			$ti, $ai, $ti_estimate_full, $ti_estimate_asymp;
	}

	if ( $opt{vvv} )
	{
		printf "(ti,ai) = (%5.2f,%5.2f) ti - ti_estimate = %13.6f - %13.6f = %13.6f\n", 
			$ti, $ai, $ti, $ti_estimate_full, $result;
	}

	my	$diff	= ($ti_estimate_full - $ti_estimate_asymp)/$ti;

	return	($result,$diff);
}
#--------------------------------------------------
sub	findTIDistribution()
{
	my	($ti,$scalar,$beta,$ai_low,$ai_high,$sfact)	= @_;

	my	$loop	= 0;
	my	$ai		= $ti;
	my	$MAXLOOP	= 100;
	my	$modeldiff= 0.0;

	while( ++$loop <= $MAXLOOP )
	{
		my	$ai_mid	= 0.5 * ($ai_low + $ai_high);
#
#		Shoot either end and in the middle to bisect.
#
		my	($low,$lowdiff)	= &generateTIResidual($ti,$ai_low,$scalar,$beta,$sfact); 
		my	($mid,$middiff)	= &generateTIResidual($ti,$ai_mid,$scalar,$beta,$sfact); 
		my	($high,$highdiff)	= &generateTIResidual($ti,$ai_high,$scalar,$beta,$sfact); 

		$modeldiff			= $middiff;

		if ( $opt{vv} )
		{
			printf "ti - ti_estimate = (%5.2f, %5.2f, %5.2f), ai = (%5.2f, %5.2f, %5.2f)\n\n", 
				$low, $mid, $high, $ai_low, $ai_mid, $ai_high;
		}
#
#		Test to see if change of sign in residual in each half.
#
		if ( $mid == 0.0 )
		{
#
#			The residual is zero at guess ai_mid, take it.
#
			$ai_low	= $ai_mid;
			$ai_high	= $ai_mid;
		}
		elsif ( $low * $mid < 0.0 )
		{
#
#			Change of sign in lower half, move top down.
#
			$ai_high	= $ai_mid;
		}
		elsif ( $mid * $high < 0.0 )
		{
#
#			Change of sign in upper half, move bottom up.
#
			$ai_low	= $ai_mid;
		}

		if ( abs($ai_high - $ai_low) < 0.001 )
		{
#
#			Near enough.
#
			$ai			= $ai_mid;
			last;
		}
	}
#
#	If we have no solution, then we return $ai = $ti.
#	This only produces acceptable plots when convergence fails
#	at the low end, however, adjustments to the bisection range
#	of the ai (aimin,aimax) have resolved these.
#
	if ( $loop > $MAXLOOP )
	{
		if ( $opt{v} )		{ print "DID NOT CONVERGE\n"; }

		$ai = $ti;
#
#		Do not apply boundary condition to asymptotic curve.
#
		if ( $opt{b} )		{ $ai = 0.0; }
	}

	return	($ai,$modeldiff);
}
#--------------------------------------------------
#	Increasing scalar moves peak towards higher ti.
#	Increasing beta shallows power-law slope for ti option (1/beta).
#	Changing them affects convergence initially at bottom and then at top.
#
if ( $opt{a} )
{
	$timax	= 10000;
}

my	$aipeak	= 0;			# peak of distribution
my	$tipeak	= 0;
my	$ailast	= 0;			# last one which converges
my	$tilast	= 0;

my	$ailow	= $aimin;
my	$aihigh	= $aimax;

my	$tilow	= $timin;
my	$tihigh	= $timax;

#
#	For both forward and inverse solutions, we must generate (ti,ai) pairs which
#	are solutions of whichever equation we are dealing with.  To get the length
#	distribution, we plot the ti on the x-axis.  For the alphabet distribution
#	we plot the ai on the x-axis.  The strategy is to crawl in from the
#	asymptote.
#
#	Solve for (ti,ai) pairs.
#
#	Generate normalising factor from furthest asymptotic value.
#
my	$sfact	= 1.0;
my	$aisum	= 0.0;
my	$asympnrm	= 0.0;				# Inverse sum of ai^(-beta)

for ( my $ti = $timax; $ti >= $timin; --$ti )
{
	if ( $opt{vv} )
	{
		print "========================================================\n";
	}

	my	$valid	= 1;

	if ( ! $opt{b} && ($aihigh > $ti) )	{ $aihigh	= $ti; }
	if ( ! $opt{b} && ($ailow > $ti) )		{ $ailow	= $aihigh - 4.0; }

	my	($ai,$diff)		= &findTIDistribution($ti,$scalar,$beta,$ailow,$aihigh,$sfact);

	if ( $ai > 0.0 )
	{
		$aisum		+= $ai;
		$asympnrm		+= power($ai,$beta,$scalar);

		push	@ai_out,	$ai;
		push	@ti_out,	$ti;
	}
	else
	{
		if ( $opt{v} )			{ printf "ti = %15.7f, ai = %15.7f, diff = %15.7f\n", $ti, $ai, $diff; }
	}
}
#
#	Traverse the output arrays, normalising under option.
#	Notice that we stop one before end so interval is defined.
#
for( my $i = 0; $i < $#ti_out; ++$i )
{
	my	$ai_l	= 0.0;			# Value of a at left of interval i.i+1
	my	$ai_u	= 0.0;			# Value of a at right of interval i.i+1

	my	$ti_l	= $ti_out[$i];
	my	$ti_u	= $ti_out[$i+1];

#	if ( $opt{n} )					{ $api	= $ai / ($scalar**(1.0/$beta)); }
	if ( $opt{n} )					{ $ai_l	= $ai_out[$i] / $aisum;	$ai_u    = $ai_out[$i+1] / $aisum; }
	else							{ $ai_l	= $ai_out[$i];			$ai_u    = $ai_out[$i+1]; }

	if ( $opt{t} )
	{
		printf "%13.7f %13.7f\n", $ti_l, $ai_l;
	}
	elsif ( $opt{a} )
	{
		printf "%13.7f %13.7f\n", $ai_l, $ti_l;
	}
	elsif ( $fixa > 0.0 )
	{
#
#		We are looking for the ti corresponding to a specific ai = fixa.
#		This will require interpolation in the interval in which it occurs.  
#		We use linear here.
#
#		** Note that to pick up both solutions we have to check for 
#		   increasing AND decreasing ai. **
#
		if ( (($fixa >= $ai_l) && ($fixa <= $ai_u))
			|| (($fixa <= $ai_l) && ($fixa >= $ai_u))	 )
		{
#
#			Linearly interpolate the corresponding value of ti.
#
			my	$ti_interp	= $ti_l + ($ti_u - $ti_l) * ($fixa - $ai_l) / ($ai_u - $ai_l);

#			printf "%13.7f %13.7f %13.7f %13.7f %13.7f %13.7f %13.7f\n", 
#				$scalar, $fixa, $ai_l, $ai_u, $ti_l, $ti_u, $ti_interp;

			printf "%13.7f\n", $ti_interp;
		}
	}
}

if ( $opt{v} )
{
	printf "Inverse normalising constant = %15.7f, scalar = %15.7f\n", $asympnrm, $scalar;
}

exit(0);
